#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022 "Space Stop" May 30, 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};


//config
#define ITERATIONS 100
#define EPSILON 1e-2
//#define EPSILON_SCALE 1.
//#define DISTANCE_STEP 1.
//#define STEP_DITHER .1
#define VOLUME_EPSILON .1
#define LIGHTING_EPSILON 1e-5

//#define LIGHT_IMPORTANCE .1
#define LIGHT_DIRECTION normalize(vec3(1))
#define LIGHT_SIZE .15
#define LIGHT_COLOR vec3(.2)
#define LIGHT_RESCALE_MAX 1e5
#define LIGHT_CAUSTICS 100.

//#define ANIMATION_TIME 10.
#define ANIMATION_SAMPLES 100
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

vec3 hue(float h) {
	return clamp(abs(mod(h*6.+vec3(0.,4.,2.),6.)-3.)-1.,0.,1.);
}

float twave(float x) {
	return abs(fract(x)-.5)*2.;
}


float box(vec3 p, vec3 s) {
	return length(max(abs(p)-s,0.));
}
float sbox(vec3 p, vec3 s) {
	p = abs(p)-s;
	return length(max(p,0.))+min(0.,max(p.x,max(p.y,p.z)));
}
float cylinder(vec3 p, vec2 s) {
	vec2 r = vec2(length(p.xz),abs(p.y))-s;
	return length(max(r,0.));
}


//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	float cs = cos(time*PI*2./10.+PI);
	return mat3(0,.6,-1.,
			0,-.4+cs*.3,mix(.3,.7*sin(time*PI*2./5.+1.4),max(0.,cs+.5)*(1./1.5)),
			1.4,0,0);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//room
	float room = sbox(p,vec3(20,6,6)+.02);
	vec2 geo = vec2(abs(room)-0.02,0.);
	
	//inside room
	if (room < 0.) {
		//table
		float td = box(p-vec3(0,-4,4),vec3(20.,2.,2.)-.2)-.2;
		if (td < geo.x) geo = vec2(td,1);
	
		//stool top
		vec3 cp = vec3(mod(p.x+200.,4.)-2.,p.y-(-3.5),p.z-1.);
		float cd = cylinder(cp,vec2(.5,.1))-.1;
		if (cd < geo.x) geo = vec2(cd,2);
	
		//stool pole
		float pd = max(length(cp.xz)-.1,abs(p.y+5.)-1.6);
		if (pd < geo.x) geo = vec2(pd,1);
		
		//tablet
		float tabd = box(p-vec3(.7,-1.98,3),vec3(.25,0,.4))-.05;
		if (tabd < geo.x) geo = vec2(tabd,4);
		
	} else {
		//space cars
		vec3 cp = vec3(p.x+time*20.,p.y-1.,p.z-20.);
		float fx = floor(cp.x/8.+1000.);
		cp.x = mod(cp.x+1000.,8.)-4.;
		//cp.y += (fract(fx*.6897234)-.5)*4.-1.;
		cp.xz = abs(cp.xz*r2d(fx*.8734));
		cp.yz = abs(cp.yz*r2d(fx*.347635));
		float car = box(cp,vec3(2,.5,.7))-fract(fx*.52731234)*.5;
		if (car < geo.x) geo = vec2(car,5.+fx);
		
	}
	
	return geo;
}
vec2 geometryDetail(vec3 p) {
	return geometry(p);
}


/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;*/
mat2x3 material(vec3 p, float mat) {
	if (mat >= 5.) {//car material
		return mat2x3(hue(mat*.786235)*.7, fract(mat*.976234)*.1,0,1);
	}
	
	if (mat == 4.) {//tablet
		if (p.y > -1.938) {//screen
			if (SFRAME%2 == 0) return mat2x3(vec3(.2), .002,0,1);//gloss
			else return mat2x3(vec3(0,1,0)*max(0.,1.-abs(mod(p.z*10.,2.)-1.-sin(time*PI*3.+p.x*floor((p.z-2.4)*10./2.)*10.))*10.), .4,1,1);
		}
		return mat2x3(vec3(.1), .06,0,1);
	}

	if (mat == 2.) {//stool
		return mat2x3(vec3(.4,.05,.04), .01,0,1);
	}
	
	if (mat == 1.) {//table/stool pole stainless steel
		return mat2x3(vec3(.8), .05,0,1);
	}

	if (p.z > 5.95) {//glass
		return mat2x3(vec3(.06,.04,.057), .0002,0,.06); 
	}
	
	if (p.y > 5.9 && abs(p.z) < 2. && mod(999.+p.x,10.) < 4.) {//roof lights
		return mat2x3(vec3(1), 1,5,1); //vec3(.9,.85,.5)
	}
	
	if (p.y < -5.5) {//tile floor
		if (SFRAME%10 == 0) return mat2x3(vec3(.2), .004,0,1);//gloss
		vec2 t2 = abs(mod(p.xz+1000.,1.)-.5);
		float t = min(1.,10.*min(t2.x,t2.y));
		return mat2x3(mix(vec3(.046,.032,.052),vec3(.34,.33,.35),t), mix(.9,.2,t),0,1);
	}

	return mat2x3(vec3(.65,.55,.4), .1,0,1);//beige wall
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	d = d*100.+200.;
	return vec3(3.7,3.4,3.9)*max(0.,(fract(dot(floor(d),vec3(.2371233,.186723,.576234)))-.9)*10.-length(fract(d)-.5)*5.);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#else
	time = 4.8;
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;

	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;
	for (int i = 0; i < ITERATIONS; i++) {
		vec2 geo = geometry(rp);
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material
			if (random.w < mat[1].z) {//stochastic opacity
				
				#ifdef LIGHT_IMPORTANCE
				if (l.w < 0.) {	//failed importance cast
					c = vec3(0);
					break;
				}
				#endif
			
				rp -= rd*epsilon*2.;//back up ray
				c *= mat[0];//apply material color
				l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
				random = hash(uvec4(abs(rp.xyz)*10000.,SFRAME))*I2F;
				vec3 nrm = normal(rp);

				#ifdef LIGHT_IMPORTANCE
				if (random.w < LIGHT_IMPORTANCE) {//directional light importance sampling
					//brdf
					float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
						brdf = clamp((dot(LIGHT_DIRECTION,reflect(rd,nrm))-edge)/(1.-min(1.,edge)),0.,1.);
					rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
					
					if (dot(rd,nrm) < -.2 || brdf < 0.) {
						c = vec3(0);
						break;
					}
					c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
					if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
					l.w = -1.;//mark ray important
					
				} else {
				#endif
				
				//randomly bounce ray off microsurface bump
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));
				#ifdef LIGHT_IMPORTANCE
				}
				#endif
				
				geo.x = 0.;
			} else {
				geo.x = VOLUME_EPSILON;
			}
		} else {
			#ifdef DISTANCE_STEP
			geo.x *= DISTANCE_STEP;
			#endif
		}
		
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.w = fract(random.w+stepSz*2763.861234);
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*(1.+LIGHT_CAUSTICS)/LIGHT_IMPORTANCE;
	} else {
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif
	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
